@@ -18,6 +18,11 @@ function(x, width = 0.9 * getOption("width"), indent = 0, exdent = 0,
             ## Format paragraph j in x[i].
             words <- z[[i]][[j]]
             nc <- nchar(words, type="w")
+	    if(any(is.na(nc))) {
+		## use byte count as a reasonable substitute
+		nc0 <- nchar(words)
+		nc[is.na(nc)] <- nc0[is.na(nc)]
+	    }
 
             ## Remove extra white space unless after a period which
             ## hopefully ends a sentence.
