@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2002  Robert Gentleman, Ross Ihaka and the
+ *  Copyright (C) 1997--2004  Robert Gentleman, Ross Ihaka and the
  *                            R Development Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -19,6 +19,11 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+/* <UTF8> char here is either ASCII or handled as a whole or as 
+   a leading portion for partial matching 
+*/
+
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -30,37 +35,47 @@
 #define SET_ARGUSED(x,v) SETLEVELS(x,v)
 
 /* Hash function and equality test for keys */
-static int K, M;
-static int(*hash) (SEXP, int);
-static int(*equal) (SEXP, int, SEXP, int);
-static SEXP HashTable;
+typedef struct _HashData HashData;
 
-static int nomatch;
+struct _HashData {
+  int K, M;
+  int(*hash) (SEXP, int, HashData *);
+  int(*equal) (SEXP, int, SEXP, int);
+  SEXP HashTable;
 
-/* Integer keys are hashed via a random number generator */
-/* based on Knuth's recommendations.  The high order K bits */
-/* are used as the hash code. */
+  int nomatch;
+};
 
-/* WARNING / FIXME : this doesn't work if K = 0 so some */
-/* fixes/warnings probably need to be installed somewhere. (RG) */
 
-static int scatter(unsigned int key)
+/* 
+   Integer keys are hashed via a random number generator
+   based on Knuth's recommendations.  The high order K bits
+   are used as the hash code.
+   
+   NB: lots of this code relies on M being a power of two and
+   on silent integer overflow mod 2^32.  It also relies on M < 31.
+
+   <FIXME>  Integer keys are wasteful for logical and raw vectors, 
+   but the tables are small in that case.
+*/
+
+static int scatter(unsigned int key, HashData *d)
 {
-    return 3141592653U * key >> (32 - K);
+    return 3141592653U * key >> (32 - d->K);
 }
 
-static int lhash(SEXP x, int indx)
+static int lhash(SEXP x, int indx, HashData *d)
 {
     if (LOGICAL(x)[indx] == NA_LOGICAL)
 	return 2;
     return LOGICAL(x)[indx];
 }
 
-static int ihash(SEXP x, int indx)
+static int ihash(SEXP x, int indx, HashData *d)
 {
     if (INTEGER(x)[indx] == NA_INTEGER)
 	return 0;
-    return scatter((unsigned int) (INTEGER(x)[indx]));
+    return scatter((unsigned int) (INTEGER(x)[indx]), d);
 }
 
 /* We use unions here because Solaris gcc -O2 has trouble with
@@ -70,52 +85,53 @@ union foo {
     unsigned int u[2];
 };
 
-static int rhash(SEXP x, int indx)
+static int rhash(SEXP x, int indx, HashData *d)
 {
     /* There is a problem with signed 0s under IEEE */
     double tmp = (REAL(x)[indx] == 0.0) ? 0.0 : REAL(x)[indx];
     /* need to use both 32-byte chunks or endianness is an issue */
-#ifdef IEEE_754 /* otherwise always false */
-    if (R_IsNaN(tmp)) tmp = R_NaN; /* we want all NaNs except NA equal */
-#endif
+    /* we want all NaNs except NA equal, and all NAs equal */
+    if (R_IsNA(tmp)) tmp = NA_REAL;
+    else if (R_IsNaN(tmp)) tmp = R_NaN;
     if (sizeof(double) >= sizeof(unsigned int)*2) {
 	union foo tmpu;
 	tmpu.d = tmp;
-	return scatter(tmpu.u[0] + tmpu.u[1]);
+	return scatter(tmpu.u[0] + tmpu.u[1], d);
     } else
-	return scatter(*((unsigned int *) (&tmp)));
+	return scatter(*((unsigned int *) (&tmp)), d);
 }
 
-static int chash(SEXP x, int indx)
+static int chash(SEXP x, int indx, HashData *d)
 {
     Rcomplex tmp;
     unsigned int u;
     tmp.r = (COMPLEX(x)[indx].r == 0.0) ? 0.0 : COMPLEX(x)[indx].r;
     tmp.i = (COMPLEX(x)[indx].i == 0.0) ? 0.0 : COMPLEX(x)[indx].i;
-#ifdef IEEE_754 /* otherwise always false */
-    if (R_IsNaN(tmp.r)) tmp.r = R_NaN; /* we want all NaNs except NA equal */
-    if (R_IsNaN(tmp.i)) tmp.i = R_NaN;
-#endif
+    /* we want all NaNs except NA equal, and all NAs equal */
+    if (R_IsNA(tmp.r)) tmp.r = NA_REAL;
+    else if (R_IsNaN(tmp.r)) tmp.r = R_NaN;
+    if (R_IsNA(tmp.i)) tmp.i = NA_REAL;
+    else if (R_IsNaN(tmp.i)) tmp.i = R_NaN;
     if (sizeof(double) >= sizeof(unsigned int)*2) {
 	union foo tmpu;
 	tmpu.d = tmp.r;
 	u = tmpu.u[0] ^ tmpu.u[1];
 	tmpu.d = tmp.i;
 	u ^= tmpu.u[0] ^ tmpu.u[1];
-	return scatter(u);
+	return scatter(u, d);
     } else
 	return scatter((*((unsigned int *)(&tmp.r)) ^
-			(*((unsigned int *)(&tmp.i)))));
+			(*((unsigned int *)(&tmp.i)))), d);
 }
 
-static int shash(SEXP x, int indx)
+static int shash(SEXP x, int indx, HashData *d)
 {
     unsigned int k;
     char *p = CHAR(STRING_ELT(x, indx));
     k = 0;
     while (*p++)
 	    k = 11 * k + *p; /* was 8 but 11 isn't a power of 2 */
-    return scatter(k);
+    return scatter(k, d);
 }
 
 static int iequal(SEXP x, int i, SEXP y, int j)
@@ -129,9 +145,7 @@ static int requal(SEXP x, int i, SEXP y, int j)
     if (!ISNAN(REAL(x)[i]) && !ISNAN(REAL(y)[j]))
 	return (REAL(x)[i] == REAL(y)[j]);
     else if (R_IsNA(REAL(x)[i]) && R_IsNA(REAL(y)[j])) return 1;
-#ifdef IEEE_754 /* otherwise always false */
     else if (R_IsNaN(REAL(x)[i]) && R_IsNaN(REAL(y)[j])) return 1;
-#endif
     else return 0;
 }
 
@@ -144,82 +158,182 @@ static int cequal(SEXP x, int i, SEXP y, int j)
     else if ((R_IsNA(COMPLEX(x)[i].r) || R_IsNA(COMPLEX(x)[i].i))
 	    && (R_IsNA(COMPLEX(y)[j].r) || R_IsNA(COMPLEX(y)[j].i)))
 	return 1;
-#ifdef IEEE_754 /* otherwise always false */
     else if ((R_IsNaN(COMPLEX(x)[i].r) || R_IsNaN(COMPLEX(x)[i].i))
 	    && (R_IsNaN(COMPLEX(y)[j].r) || R_IsNaN(COMPLEX(y)[j].i)))
 	return 1;
-#endif
-    else 
+    else
 	return 0;
 }
 
 static int sequal(SEXP x, int i, SEXP y, int j)
 {
-    if (STRING_ELT(x, i) != NA_STRING && STRING_ELT(y, j) != NA_STRING)
-	return !strcmp(CHAR(STRING_ELT(x, i)), CHAR(STRING_ELT(y, j)));
-    else 
-	return STRING_ELT(x, i) == STRING_ELT(y, j);
+    /* Two strings which have the same address must be the same, 
+       so avoid looking at the contents */
+    if (STRING_ELT(x, i) == STRING_ELT(y, j)) return 1;
+    /* Then if either is NA the other cannot be */
+    if (STRING_ELT(x, i) == NA_STRING || STRING_ELT(y, j) == NA_STRING) 
+	return 0;
+    /* Finally look at the contents if necessary */
+    return !strcmp(CHAR(STRING_ELT(x, i)), CHAR(STRING_ELT(y, j)));
 }
 
-/* Choose M to be the smallest power of 2 */
-/* not less than 4*n and set K = log2(M) */
-static void MKsetup(int n)
+static int rawhash(SEXP x, int indx, HashData *d)
 {
-    int n4 = 4 * n;
-    M = 1;
-    K = 0;
-    while (M < n4) {
-	M *= 2;
-	K += 1;
+    return RAW(x)[indx];
+}
+
+static int rawequal(SEXP x, int i, SEXP y, int j)
+{
+    return (RAW(x)[i] == RAW(y)[j]);
+}
+
+static int vhash(SEXP x, int indx, HashData *d)
+{
+    int i;
+    unsigned int key;
+    SEXP this = VECTOR_ELT(x, indx);
+    
+    key = OBJECT(this) + 2*TYPEOF(this) + 100*length(this);
+    /* maybe we should also look at attributes, but that slows us down */
+    switch (TYPEOF(this)) {
+    case LGLSXP:
+	/* This is not too clever: pack into 32-bits and then scatter? */
+	for(i = 0; i < LENGTH(this); i++) {
+	    key ^= lhash(this, i, d);
+	    key *= 97;
+	}
+	break;
+    case INTSXP:
+	for(i = 0; i < LENGTH(this); i++) {
+	    key ^= ihash(this, i, d);
+	    key *= 97;
+	}
+	break;
+    case REALSXP:
+	for(i = 0; i < LENGTH(this); i++) {
+	    key ^= rhash(this, i, d);
+	    key *= 97;
+	}
+	break;
+    case CPLXSXP:
+	for(i = 0; i < LENGTH(this); i++) {
+	    key ^= chash(this, i, d);
+	    key *= 97;
+	}
+	break;
+    case STRSXP:
+	for(i = 0; i < LENGTH(this); i++) {
+	    key ^= shash(this, i, d);
+	    key *= 97;
+	}
+	break;
+    case RAWSXP:
+	for(i = 0; i < LENGTH(this); i++) {
+	    key ^= scatter(rawhash(this, i, d), d);
+	    key *= 97;
+	}
+	break;
+    case VECSXP:
+	for(i = 0; i < LENGTH(this); i++) {
+	    key ^= vhash(this, i, d);
+	    key *= 97;
+	}
+	break;
+    default:
+	break;
+    }
+    return scatter(key, d);
+}
+
+Rboolean compute_identical(SEXP x, SEXP y); /* from identical.c */
+
+static int vequal(SEXP x, int i, SEXP y, int j)
+{
+    return compute_identical(VECTOR_ELT(x, i), VECTOR_ELT(y, j));
+}
+
+/*
+  Choose M to be the smallest power of 2
+  not less than 2*n and set K = log2(M).
+  Need K >= 1 and hence M >= 2, and 2^M <= 2^31 -1, hence n <= 2^29.
+
+  Dec 2004: modified from 4*n to 2*n, since in the worst case we have
+  a 50% full table, and that is still rather efficient -- see
+  R. Sedgewick (1998) Algorithms in C++ 3rd edition p.606.
+*/
+static void MKsetup(int n, HashData *d)
+{
+    int n4 = 2 * n;
+
+    if(n < 0 || n > 536870912) /* protect against overflow to -ve */
+	error(_("length %d is too large for hashing"), n);
+    d->M = 2;
+    d->K = 1;
+    while (d->M < n4) {
+	d->M *= 2;
+	d->K += 1;
     }
 }
 
-static void HashTableSetup(SEXP x)
+static void HashTableSetup(SEXP x, HashData *d)
 {
     switch (TYPEOF(x)) {
     case LGLSXP:
-	hash = lhash;
-	equal = iequal;
-	MKsetup(3);
+	d->hash = lhash;
+	d->equal = iequal;
+	MKsetup(3, d);
 	break;
     case INTSXP:
-	hash = ihash;
-	equal = iequal;
-	MKsetup(LENGTH(x));
+	d->hash = ihash;
+	d->equal = iequal;
+	MKsetup(LENGTH(x), d);
 	break;
     case REALSXP:
-	hash = rhash;
-	equal = requal;
-	MKsetup(LENGTH(x));
+	d->hash = rhash;
+	d->equal = requal;
+	MKsetup(LENGTH(x), d);
 	break;
     case CPLXSXP:
-	hash = chash;
-	equal = cequal;
-	MKsetup(LENGTH(x));
+	d->hash = chash;
+	d->equal = cequal;
+	MKsetup(LENGTH(x), d);
 	break;
     case STRSXP:
-	hash = shash;
-	equal = sequal;
-	MKsetup(LENGTH(x));
+	d->hash = shash;
+	d->equal = sequal;
+	MKsetup(LENGTH(x), d);
 	break;
+    case RAWSXP:
+	d->hash = rawhash;
+	d->equal = rawequal;
+	d->M = 256;
+	d->K = 8; /* unused */
+	break;
+    case VECSXP:
+	d->hash = vhash;
+	d->equal = vequal;
+	MKsetup(LENGTH(x), d);
+	break;
+    default:
+	UNIMPLEMENTED_TYPE("HashTableSetup", x);
     }
-    HashTable = allocVector(INTSXP, M);
+    d->HashTable = allocVector(INTSXP, d->M);
 }
 
 /* Open address hashing */
 /* Collision resolution is by linear probing */
 /* The table is guaranteed large so this is sufficient */
 
-static int isDuplicated(SEXP x, int indx)
+static int isDuplicated(SEXP x, int indx, HashData *d)
 {
     int i, *h;
 
-    h = INTEGER(HashTable);
-    i = hash(x, indx);
+    h = INTEGER(d->HashTable);
+    i = d->hash(x, indx, d);
     while (h[i] != NIL) {
-	if (equal(x, h[i], x, indx))
+	if (d->equal(x, h[i], x, indx))
 	    return 1;
-	i = (i + 1) % M;
+	i = (i + 1) % d->M;
     }
     h[i] = indx;
     return 0;
@@ -230,26 +344,30 @@ SEXP duplicated(SEXP x)
     SEXP ans;
     int *h, *v;
     int i, n;
+    HashData data;
+
+    if (!isVector(x)) 
+	error(_("'duplicated' applies only to vectors"));
 
     n = LENGTH(x);
-    HashTableSetup(x);
-    PROTECT(HashTable);
+    HashTableSetup(x, &data);
+    PROTECT(data.HashTable);
     ans = allocVector(LGLSXP, n);
     UNPROTECT(1);
-    h = INTEGER(HashTable);
+    h = INTEGER(data.HashTable);
     v = LOGICAL(ans);
 
-    for (i = 0; i < M; i++)
+    for (i = 0; i < data.M; i++)
 	h[i] = NIL;
 
     for (i = 0; i < n; i++)
-	v[i] = isDuplicated(x, i);
+	v[i] = isDuplicated(x, i, &data);
 
     return ans;
 }
 
-/* .Internal(duplicated(x)) [op=0]  and
-   .Internal(unique(x))	    [op=1] :
+/* .Internal(duplicated(x))       [op=0]
+   .Internal(unique(x))	          [op=1]
 */
 SEXP do_duplicated(SEXP call, SEXP op, SEXP args, SEXP env)
 {
@@ -258,20 +376,21 @@ SEXP do_duplicated(SEXP call, SEXP op, SEXP args, SEXP env)
 
     checkArity(op, args);
     x = CAR(args);
-    /* handle zero length vectors */
-    if (length(x) == 0)
-	return(allocVector(PRIMVAL(op) == 0 ? LGLSXP : TYPEOF(x), 0));
+    /* handle zero length vectors, and NULL */
+    if ((n = length(x)) == 0)
+	return(allocVector(PRIMVAL(op) != 1 ? LGLSXP : TYPEOF(x), 0));
 
-    if (!(isVectorAtomic(x)))
-	error("%s() applies only to vectors",
+    if (!isVector(x)) {
+	PrintValue(x);
+	error(_("%s() applies only to vectors"),
 	      (PRIMVAL(op) == 0 ? "duplicated" : "unique"));
+    }
 
     dup = duplicated(x);
     if (PRIMVAL(op) == 0) /* "duplicated()" : */
 	return dup;
     /*	ELSE
 	use the results of "duplicated" to get "unique" */
-    n = LENGTH(x);
 
     /* count unique entries */
     k = 0;
@@ -309,41 +428,53 @@ SEXP do_duplicated(SEXP call, SEXP op, SEXP args, SEXP env)
 	    if (LOGICAL(dup)[i] == 0)
 		SET_STRING_ELT(ans, k++, STRING_ELT(x, i));
 	break;
+    case VECSXP:
+	for (i = 0; i < n; i++)
+	    if (LOGICAL(dup)[i] == 0)
+		SET_VECTOR_ELT(ans, k++, VECTOR_ELT(x, i));
+	break;
+    case RAWSXP:
+	for (i = 0; i < n; i++)
+	    if (LOGICAL(dup)[i] == 0)
+		RAW(ans)[k++] = RAW(x)[i];
+	break;
+    default:
+	UNIMPLEMENTED_TYPE("duplicated", x);
     }
     return ans;
 }
 
 /* Build a hash table, ignoring information on duplication */
-static void DoHashing(SEXP table)
+static void DoHashing(SEXP table, HashData *d)
 {
     int *h, i, n;
 
     n = LENGTH(table);
-    h = INTEGER(HashTable);
+    h = INTEGER(d->HashTable);
 
-    for (i = 0; i < M; i++)
+    for (i = 0; i < d->M; i++)
 	h[i] = NIL;
 
     for (i = 0; i < n; i++)
-	(void) isDuplicated(table, i);
+	(void) isDuplicated(table, i, d);
 }
 
-static int Lookup(SEXP table, SEXP x, int indx)
+static int Lookup(SEXP table, SEXP x, int indx, HashData *d)
 {
     int i, *h;
 
-    h = INTEGER(HashTable);
-    i = hash(x, indx);
+    h = INTEGER(d->HashTable);
+    i = d->hash(x, indx, d);
     while (h[i] != NIL) {
-	if (equal(table, h[i], x, indx))
+	if (d->equal(table, h[i], x, indx))
 	    return h[i] + 1;
-	i = (i + 1) % M;
+	i = (i + 1) % d->M;
     }
-    return nomatch;
+    return d->nomatch;
 }
 
 /* Now do the table lookup */
-static SEXP HashLookup(SEXP table, SEXP x)
+static SEXP HashLookup(SEXP table, SEXP x, HashData *d)
 {
     SEXP ans;
     int i, n;
@@ -351,27 +482,30 @@ static SEXP HashLookup(SEXP table, SEXP x)
     n = LENGTH(x);
     ans = allocVector(INTSXP, n);
     for (i = 0; i < n; i++) {
-	INTEGER(ans)[i] = Lookup(table, x, i);
+	INTEGER(ans)[i] = Lookup(table, x, i, d);
     }
     return ans;
 }
 
 SEXP do_match(SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    SEXP x, table, ans;
+#ifdef NOTMOVED
+    SEXP x, table;
     SEXPTYPE type;
-    int n, i;
+#endif
+    int nomatch;
 
     checkArity(op, args);
 
     if ((!isVector(CAR(args)) && !isNull(CAR(args)))
 	|| (!isVector(CADR(args)) && !isNull(CADR(args))))
-	error("match requires vector arguments");
+	error(_("'match' requires vector arguments"));
 
     /* Coerce to a common type; type == NILSXP is ok here.
      * Note that R's match() does only coerce factors (to character).
      * Hence, coerce to character or to `higher' type
      * (given that we have "Vector" or NULL) */
+#ifdef NOTMOVED
     if(TYPEOF(CAR(args))  >= STRSXP || TYPEOF(CADR(args)) >= STRSXP)
 	type = STRSXP;
     else type = TYPEOF(CAR(args)) < TYPEOF(CADR(args)) ?
@@ -379,35 +513,49 @@ SEXP do_match(SEXP call, SEXP op, SEXP args, SEXP env)
     x = SETCAR(args, coerceVector(CAR(args), type));
     table = SETCADR(args, coerceVector(CADR(args), type));
     nomatch = asInteger(CAR(CDDR(args)));
-    n = length(x);
+    return match(table, x, nomatch);
+#else
+    nomatch = asInteger(CAR(CDDR(args)));
+    return match(CADR(args), CAR(args), nomatch);
+#endif
+}
 
-    /* handle zero length arrays */
-    if (n == 0) return allocVector(INTSXP, 0);
+SEXP match(SEXP itable, SEXP ix, int nmatch)
+{
+    SEXP ans, x, table;
+    SEXPTYPE type;
+    HashData data;
+    int i, n = length(ix);
+
+    /* Coerce to a common type; type == NILSXP is ok here.
+     * Note that R's match() does only coerce factors (to character).
+     * Hence, coerce to character or to `higher' type
+     * (given that we have "Vector" or NULL) */
+    if(TYPEOF(ix)  >= STRSXP || TYPEOF(itable) >= STRSXP)
+        type = STRSXP;
+    else type = TYPEOF(ix) < TYPEOF(itable) ?
+             TYPEOF(itable) : TYPEOF(ix);
+    PROTECT(x = coerceVector(ix, type));
+    PROTECT(table = coerceVector(itable, type));
+
+    /* handle zero length arguments */
+    if (n == 0) {
+	UNPROTECT(2);
+	return allocVector(INTSXP, 0);
+    }
     if (length(table) == 0) {
 	ans = allocVector(INTSXP, n);
 	for (i = 0; i < n; i++)
-	    INTEGER(ans)[i] = nomatch;
+	    INTEGER(ans)[i] = nmatch;
+	UNPROTECT(2);
 	return ans;
     }
-
-    HashTableSetup(table);
-    PROTECT(HashTable);
-    DoHashing(table);
-    ans = HashLookup(table, x);
-    UNPROTECT(1);
-    return ans;
-}
-
-SEXP match(SEXP table, SEXP x, int nmatch)
-{
-    SEXP ans;
-
-    nomatch=nmatch;
-    HashTableSetup(table);
-    PROTECT(HashTable);
-    DoHashing(table);
-    ans = HashLookup(table, x);
-    UNPROTECT(1);
+    data.nomatch = nmatch;
+    HashTableSetup(table, &data);
+    PROTECT(data.HashTable);
+    DoHashing(table, &data);
+    ans = HashLookup(table, x, &data);
+    UNPROTECT(3);
     return ans;
 }
 
@@ -437,10 +585,10 @@ SEXP do_pmatch(SEXP call, SEXP op, SEXP args, SEXP env)
     n_target = LENGTH(target);
     dups_ok = asLogical(CADDR(args));
     if (dups_ok == NA_LOGICAL)
-	errorcall(call, "invalid \"duplicates.ok\" argument");
+	errorcall(call, _("invalid 'duplicates.ok' argument"));
 
     if (!isString(input) || !isString(target))
-	errorcall(call, "argument is not of mode character");
+	errorcall(call, _("argument is not of mode character"));
 
     used = (int *) R_alloc(n_target, sizeof(int));
     for (j = 0; j < n_target; j++) used[j] = 0;
@@ -484,32 +632,6 @@ SEXP do_pmatch(SEXP call, SEXP op, SEXP args, SEXP env)
 	}
     }
 
-#ifdef OLD_PMATCH
-    for (i = 0; i < n_input; i++) {
-	temp = strlen(CHAR(STRING_ELT(input, i)));
-	mtch = 0;
-	mtch_count = 0;
-	if (temp) {
-	    for (j = 0; j < n_target; j++) {
-		k = strncmp(CHAR(STRING_ELT(input, i)),
-			    CHAR(STRING_ELT(target, j)), temp);
-		if (k == 0) {
-		    mtch = j + 1;
-		    if (dups_ok ||
-			strlen(CHAR(STRING_ELT(target, j))) == temp)
-			/* This is odd, effectively sets dups.ok
-			 * for perfect mtches, but that's what
-			 * Splus 3.4 does  --pd
-			 */
-			break;
-		    if (mtch_count++ && !dups_ok)
-			mtch = 0;
-		}
-	    }
-	}
-	INTEGER(ans)[i] = mtch;
-    }
-#endif
     vmaxset(vmax);
     return ans;
 }
@@ -532,7 +654,7 @@ SEXP do_charmatch(SEXP call, SEXP op, SEXP args, SEXP env)
     n_target = LENGTH(target);
 
     if (!isString(input) || !isString(target))
-	errorcall(call, "argument is not of mode character");
+	errorcall(call, _("argument is not of mode character"));
 
     ans = allocVector(INTSXP, n_input);
 
@@ -620,7 +742,7 @@ static SEXP subDots(SEXP rho)
     dots = findVar(R_DotsSymbol, rho);
 
     if (dots == R_UnboundValue)
-	error("... used in a situation where it doesn't exist");
+	error(_("... used in a situation where it doesn't exist"));
 
     if (dots == R_MissingArg)
 	return dots;
@@ -658,8 +780,8 @@ SEXP do_matchcall(SEXP call, SEXP op, SEXP args, SEXP env)
 	funcall = VECTOR_ELT(funcall, 0);
 
     if (TYPEOF(funcall) != LANGSXP) {
-	b = deparse1(funcall, 1);
-	errorcall(call, "%s is not a valid call", CHAR(STRING_ELT(b, 0)));
+	b = deparse1(funcall, 1, SIMPLEDEPARSE);
+	errorcall(call, _("'%s' is not a valid call"), CHAR(STRING_ELT(b, 0)));
     }
 
     /* Get the function definition */
@@ -678,7 +800,34 @@ SEXP do_matchcall(SEXP call, SEXP op, SEXP args, SEXP env)
 	    sysp = R_GlobalEnv;
 	else
 	    sysp = cptr->sysparent;
-	if ( TYPEOF(CAR(funcall)) == SYMSXP )
+	if (cptr != NULL)
+	    /* Changed to use the function from which match.call was
+	       called as recorded in the context.  This change is
+	       needed in case the current function is computed in a
+	       way that cannot be reproduced by a second computation,
+	       or if it is a registered S3 method that is not
+	       lexically visible at the call site.
+
+	       There is one particular case where this represents a
+	       change from previous semantics: The definition is NULL,
+	       the call is supplied explicitly, and the function in
+	       the call is NOT the current function.  The new behavior
+	       is to ignore the function in the call and use the
+	       current function.  This is consistent with (my reading
+	       of) the documentation in both R and Splus.  However,
+	       the old behavior of R was consistent with the behavior
+	       of Splus (and inconsistent with the documentation in
+	       both cases).
+
+	       The previous semantics for this case can be restored by
+	       having the .Internal receive an additional argument
+	       that indicates whether the call was supplied explicitly
+	       or missing, and using the function recorded in the
+	       context only if the call was not supplied explicitly.
+	       The documentation should also be changed to be
+	       consistent with this behavior.  LT */
+	    PROTECT(b = duplicate(cptr->callfun));
+	else if ( TYPEOF(CAR(funcall)) == SYMSXP )
 	    PROTECT(b = findFun(CAR(funcall), sysp));
 	else
 	    PROTECT(b = eval(CAR(funcall), sysp));
@@ -688,16 +837,16 @@ SEXP do_matchcall(SEXP call, SEXP op, SEXP args, SEXP env)
     /* It must be a closure! */
 
     if (TYPEOF(b) != CLOSXP) {
-	b = deparse1(b, 1);
-	errorcall(call, "%s is not a function", CHAR(STRING_ELT(b, 0)));
+	b = deparse1(b, 1, SIMPLEDEPARSE);
+	errorcall(call, _("'%s' is not a function"), CHAR(STRING_ELT(b, 0)));
     }
 
     /* Do we expand ... ? */
 
     expdots = asLogical(CAR(CDDR(args)));
     if (expdots == NA_LOGICAL) {
-	b = deparse1(CADDR(args), 1);
-	errorcall(call, "%s is not a logical", CHAR(STRING_ELT(b, 0)));
+	b = deparse1(CADDR(args), 1, SIMPLEDEPARSE);
+	errorcall(call, _("'%s' is not a logical"), CHAR(STRING_ELT(b, 0)));
     }
 
     /* Get the formals and match the actual args */
@@ -774,3 +923,264 @@ SEXP do_matchcall(SEXP call, SEXP op, SEXP args, SEXP env)
     return rval;
 }
 
+
+#if defined(HAVE_STRING_H)
+#  include <string.h>
+#  ifdef _AIX  /*some people just have to be different */
+#    include <memory.h>
+#  endif
+/* int and double zeros are all bits off */
+#  define ZEROINT(X,N,I) do{memset(INTEGER(X),0,N*sizeof(int));}while(0)
+#  define ZERODBL(X,N,I) do{memset(REAL(X),0,N*sizeof(double));}while(0)
+#else
+#  define ZEROINT(X,N,I) for(I=0;I<N;I++) INTEGER(X)[I]=0
+#  define ZERODBL(X,N,I) for(I=0;I<N;I++) REAL(X)[I]=0
+#endif
+
+SEXP Rrowsum_matrix(SEXP x, SEXP ncol, SEXP g, SEXP uniqueg)
+{
+    SEXP matches,ans;
+    int i, j, n, p,ng=0,offset,offsetg;
+    HashData data;
+    data.nomatch = 0;
+
+    n = LENGTH(g);
+    p= INTEGER(ncol)[0];
+    ng=length(uniqueg);
+
+    HashTableSetup(uniqueg, &data);
+    PROTECT(data.HashTable);
+    DoHashing(uniqueg, &data);
+    PROTECT(matches = HashLookup(uniqueg ,g, &data));
+
+    PROTECT(ans=allocMatrix(TYPEOF(x),ng,p));
+
+    offset=0; offsetg=0;
+
+    switch(TYPEOF(x)){
+    case REALSXP:
+	ZERODBL(ans,ng*p,i);
+	for(i=0;i<p;i++){
+	    for(j=0;j<n;j++){
+		REAL(ans)[INTEGER(matches)[j]-1+offsetg]+=REAL(x)[j+offset];
+	    }
+	    offset+=n;
+	    offsetg+=ng;
+	}
+	break;
+    case INTSXP:
+	ZEROINT(ans,ng*p,i);
+	for(i=0;i<p;i++){
+	    for(j=0;j<n;j++){
+		if (INTEGER(x)[j+offset]==NA_INTEGER)
+		    INTEGER(ans)[INTEGER(matches)[j]-1+offsetg]=NA_INTEGER;
+		else if  (INTEGER(ans)[INTEGER(matches)[j]-1+offsetg]!=NA_INTEGER)
+		    INTEGER(ans)[INTEGER(matches)[j]-1+offsetg]+=INTEGER(x)[j+offset];
+	    }
+	    offset+=n;
+	    offsetg+=ng;
+	}
+	break;
+    default:
+	error(_("non-numeric matrix in rowsum(): this cannot happen"));
+    }
+
+    UNPROTECT(2); /*HashTable, matches*/
+    UNPROTECT(1); /*ans*/
+    return ans;
+}
+
+SEXP Rrowsum_df(SEXP x, SEXP ncol, SEXP g, SEXP uniqueg)
+{
+    SEXP matches,ans,col,xcol;
+    int i, j, n, p,ng=0,offset,offsetg;
+    HashData data;
+    data.nomatch = 0;
+
+    n = LENGTH(g);
+    p= INTEGER(ncol)[0];
+    ng=length(uniqueg);
+
+    HashTableSetup(uniqueg, &data);
+    PROTECT(data.HashTable);
+    DoHashing(uniqueg, &data);
+    PROTECT(matches = HashLookup(uniqueg ,g, &data));
+
+    PROTECT(ans=allocVector(VECSXP,p));
+
+    offset=0; offsetg=0;
+
+    for(i=0; i<p;i++){
+	xcol=VECTOR_ELT(x,i);
+	if (!isNumeric(xcol))
+	    error(_("non-numeric data frame in rowsum"));
+	switch(TYPEOF(xcol)){
+	case REALSXP:
+	    PROTECT(col=allocVector(REALSXP,ng));
+	    ZERODBL(col,ng,i);
+	    for(j=0;j<n;j++){
+		REAL(col)[INTEGER(matches)[j]-1]+=REAL(xcol)[j];
+	    }
+	    SET_VECTOR_ELT(ans,i,col);
+	    UNPROTECT(1);
+	    break;
+	case INTSXP:
+	    PROTECT(col=allocVector(INTSXP,ng));
+	    ZEROINT(col,ng,i);
+	    for(j=0;j<n;j++){
+		if (INTEGER(xcol)[j]==NA_INTEGER)
+		    INTEGER(col)[INTEGER(matches)[j]-1]=NA_INTEGER;
+		else if (INTEGER(col)[INTEGER(matches)[j]-1]!=NA_INTEGER)
+		    INTEGER(col)[INTEGER(matches)[j]-1]+=INTEGER(xcol)[j];
+	    }
+	    SET_VECTOR_ELT(ans,i,col);
+	    UNPROTECT(1);
+	    break;
+
+	default:
+	    error(_("this cannot happen"));
+	}
+    }
+    namesgets(ans,getAttrib(x,R_NamesSymbol));
+
+    UNPROTECT(2); /*HashTable, matches*/
+    UNPROTECT(1); /*ans*/
+    return ans;
+}
+
+/* returns 1-based duplicate no */
+static int isDuplicated2(SEXP x, int indx, HashData *d)
+{
+    int i, *h;
+
+    h = INTEGER(d->HashTable);
+    i = d->hash(x, indx, d);
+    while (h[i] != NIL) {
+	if (d->equal(x, h[i], x, indx))
+	    return h[i] + 1;
+	i = (i + 1) % d->M;
+    }
+    h[i] = indx;
+    return 0;
+}
+
+static SEXP duplicated2(SEXP x, HashData *d)
+{
+    SEXP ans;
+    int *h, *v;
+    int i, n;
+
+    n = LENGTH(x);
+    HashTableSetup(x, d);
+    PROTECT(d->HashTable);
+    ans = allocVector(INTSXP, n);
+    UNPROTECT(1);
+    h = INTEGER(d->HashTable);
+    v = INTEGER(ans);
+    for (i = 0; i < d->M; i++) h[i] = NIL;
+    for (i = 0; i < n; i++) v[i] = isDuplicated2(x, i, d);
+    return ans;
+}
+
+#include <R_ext/RS.h> /* for Calloc, free */
+
+SEXP do_makeunique(SEXP call, SEXP op, SEXP args, SEXP env)
+{
+    SEXP names, sep, ans, dup, newx;
+    int i, n, cnt, len, maxlen = 0, *cnts, dp;
+    HashData data;
+    char *csep, *buf;
+
+    checkArity(op, args);
+    names = CAR(args);
+    if(!isString(names))
+	errorcall(call, _("'names' must be a character vector"));
+    n = LENGTH(names);
+    sep = CADR(args);
+    if(!isString(sep) || LENGTH(sep) != 1)
+	errorcall(call, _("'sep' must be a character string"));
+    csep = CHAR(STRING_ELT(sep, 0));
+    PROTECT(ans = allocVector(STRSXP, n));
+    for(i = 0; i < n; i++) {
+	SET_STRING_ELT(ans, i, STRING_ELT(names, i));
+	len = strlen(CHAR(STRING_ELT(names, i)));
+	if(len > maxlen) maxlen = len;
+    }
+    if(n > 1) {
+	/* +2 for terminator and rounding error */
+	buf = Calloc(maxlen + strlen(csep) + log((double)n)/log(10.0) + 2,
+		     char);
+	cnts = Calloc(n, int);
+	for(i = 0; i < n; i++) cnts[i] = 1;
+	data.nomatch = 0;
+	PROTECT(newx = allocVector(STRSXP, 1));
+	PROTECT(dup = duplicated2(names, &data));
+	PROTECT(data.HashTable);
+	for(i = 1; i < n; i++) { /* first cannot be a duplicate */
+	    dp = INTEGER(dup)[i]; /* 1-based number of first occurrence */
+	    if(dp == 0) continue;
+	    /* Try appending 1,2,3, ..., n-1 until it is not already in use */
+	    for(cnt = cnts[dp-1]; cnt < n; cnt++) {
+		sprintf(buf, "%s%s%d", CHAR(STRING_ELT(names, i)), csep, cnt);
+		SET_STRING_ELT(newx, 0, mkChar(buf));
+		if(Lookup(ans, newx, 0, &data) == data.nomatch) break;
+	    }
+	    SET_STRING_ELT(ans, i, STRING_ELT(newx, 0));
+	    /* insert it */ (void) isDuplicated(ans, i, &data);
+	    cnts[dp-1] = cnt+1; /* cache the first unused cnt value */
+	}
+	Free(cnts);
+	Free(buf);
+	UNPROTECT(3);
+    }
+    UNPROTECT(1);
+    return ans;
+}
+
+/* Use hashing to improve object.size. Here we want equal CHARSXPs, 
+   not equal contents.  This only uses the bottom 32 bits of the pointer, 
+   but for now that's almost certainly OK */
+
+static int cshash(SEXP x, int indx, HashData *d)
+{
+    return scatter((unsigned int) STRING_ELT(x, indx), d);
+}
+
+static int csequal(SEXP x, int i, SEXP y, int j)
+{
+    return STRING_ELT(x, i) == STRING_ELT(y, j);
+}
+
+static void HashTableSetup1(SEXP x, HashData *d)
+{
+    d->hash = cshash;
+    d->equal = csequal;
+    MKsetup(LENGTH(x), d);
+    d->HashTable = allocVector(INTSXP, d->M);
+}
+
+SEXP csduplicated(SEXP x)
+{
+    SEXP ans;
+    int *h, *v;
+    int i, n;
+    HashData data;
+
+    if(TYPEOF(x) != STRSXP)
+	error(_("csduplicated not called on a STRSXP"));
+    n = LENGTH(x);
+    HashTableSetup1(x, &data);
+    PROTECT(data.HashTable);
+    ans = allocVector(LGLSXP, n);
+    UNPROTECT(1);
+    h = INTEGER(data.HashTable);
+    v = LOGICAL(ans);
+
+    for (i = 0; i < data.M; i++)
+	h[i] = NIL;
+
+    for (i = 0; i < n; i++)
+	v[i] = isDuplicated(x, i, &data);
+
+    return ans;
+}
